import pytest
import json
import sys
import os

from eth_tools.transfer_event_parser import TransferEventParser

mock_event = {"args": {"from": "0xA8502b841D9fcD5d2a4C7439A13e87796AEd56A1", "to": "0x0001FB050Fe7312791bF6475b96569D83F695C9f", "value": 174003773318124380}, "event": "Transfer", "logIndex": 44, "transactionIndex": 40,
              "transactionHash": "0x69dd5baf185d9c4349d70a84963885ed301614d3c8d0457955d4df06e8bcbcb8", "address": "0x0bc529c00C6401aEF6D220BE8C6Ea1667F6Ad93e", "blockHash": "0x858a94be7abaa2aa124a3aef22bd4355afd460164b63b400ea12af7a3128e90f", "blockNumber": 10478438}


@pytest.fixture
def addresses():
    fname = os.path.join(os.path.dirname(__file__),
                         'test_data/dummyAddresses.json')
    with open(fname) as f:
        addresses = json.load(f)
    return addresses


@pytest.fixture
def events():
    fname = os.path.join(os.path.dirname(__file__),
                         'test_data/dummyEvents.json')
    with open(fname) as f:
        events = [json.loads(e) for e in f]
    return events


@pytest.fixture
def dummy_balances():
    return {"0x123": {2: 10, 3: 11, 5: 18}}


@pytest.fixture
def dummy_addresses():
    return {"addr": "0x123"}


def test_handle_transfer_event(addresses):
    transfer_event_parser = TransferEventParser(addresses)
    transfer_event_parser.set_keys(mock_event)
    transfer_event_parser.handle_transfer_event(mock_event)
    assert transfer_event_parser.balances["0x0001FB050Fe7312791bF6475b96569D83F695C9f"][10478438] == 174003773318124380


def test_execute_events(addresses, events):
    transfer_event_parser = TransferEventParser(addresses)
    transfer_event_parser.execute_events(events)
    assert transfer_event_parser.balances[
        "0x0001FB050Fe7312791bF6475b96569D83F695C9f"][10478649] == 9995796750644777900207
    assert transfer_event_parser.balances["0x28b88cfD875C883cDb61938C97B8d1baabf31c88"][10478826] == 90896908769421890


def test_write_balances(dummy_addresses, dummy_balances, tmp_path):
    transfer_event_parser = TransferEventParser(dummy_addresses)
    transfer_event_parser.balances = dummy_balances
    transfer_event_parser.start = 2
    transfer_event_parser.write_balances("tok", filepath=str(tmp_path) + "/")
    output_path = tmp_path / "tok-balances:addr.csv"
    assert os.path.exists(output_path)
    with open(output_path) as f:
        print(f.read())
    with open(output_path) as f:
        balances = [json.loads(line) for line in f]
    assert balances == [{"blockNumber": 2, "balance": 10}, {"blockNumber": 3, "balance": 11}, {
        "blockNumber": 4, "balance": 11}, {"blockNumber": 5, "balance": 18}]



import os
import json

from collections import defaultdict
from eth_tools.logger import logger


class TransferEventParser:
    """Parses ERC20 'transfer' events"""

    def __init__(self, addresses: dict, start: int = None, end: int = None):
        self.addresses = addresses
        for key, val in self.addresses.items():
            self.addresses[key] = val
        self.balances = defaultdict(lambda: defaultdict(lambda: 0))
        self.last_update = defaultdict(lambda: 0)
        self.start = start
        self.end = end
        self.keys = {}
        self.keys['amount'] = 'amount'
        self.keys['to'] = 'to'
        self.keys['from'] = 'from'

    def execute_events(self, events: list):
        first_event = True
        for event in events:
            if event['event'] == 'Transfer':
                if first_event:
                    self.set_keys(event)
                    self.start = self.start if self.start is not None else event['blockNumber']
                    first_event = False
                self.handle_transfer_event(event)

    def handle_transfer_event(self, event: dict):
        if event['event'] != 'Transfer':
            logger.warning('Failed to parse non-transfer event')
            return
        key_to = self.keys['to']
        key_from = self.keys['from']
        key_amount = self.keys['amount']
        block = event['blockNumber']
        if event['args'][key_from] in self.addresses.values():
            self.update_balance(event['args'][key_from],
                                -1 * event['args'][key_amount], block)
        if event['args'][key_to] in self.addresses.values():
            self.update_balance(event['args'][key_to],
                                event['args'][key_amount], block)

    def update_balance(self, account: str, change: int, block: int):
        last_updated_block = self.last_update[account]
        self.balances[account][block] = self.balances[account][last_updated_block] + change
        self.last_update[account] = block

    def find_key(self, event: dict, candidates: set):
        key = next(iter(candidates & event.keys()), None)
        if key:
            return key
        raise ValueError(f"none of {candidates} found in {event}")

    def set_keys(self, event: dict) -> bool:
        """ERC20 contracts don't follow a standard name for transfer args.
        This should be handled here."""
        self.keys["from"] = self.find_key(event['args'], {"from", "_from"})
        self.keys["to"] = self.find_key(event['args'], {"to", "_to"})
        self.keys["amount"] = self.find_key(
            event['args'], {"amount", "value", "_value"})

    def write_balances(self, token: str, interval: int = None, filepath: str = None):
        for name, address in self.addresses.items():
            fname = token.lower()+"-balances:"+name.lower()+'.csv'
            if filepath is not None:
                fname = filepath + fname
            with open(fname, "w") as f:
                self.write_address_balances(address, f, interval)

    def write_address_balances(self, address, f, interval: int = None):
        first_block = True
        last_balance = 0
        inter = 1 if interval is None else interval
        counter = 0
        for block in self.balances[address].keys():
            if first_block:
                last_block = block
                first_block = False
            if block == 0:
                continue
            # fill in missing blocks
            block_number = last_block + 1
            while block_number <= block - 1:
                counter += 1
                if counter % inter == 0:
                    self.log_balance(
                        counter, f, block_number, last_balance)
                block_number += 1
            counter += 1
            if counter % inter == 0:
                self.log_balance(
                    counter, f, block, self.balances[address][block])
            last_balance = self.balances[address][block]
            last_block = block

    def log_balance(self, counter: int, f, block_number: int, balance: int):
        if self.start is not None and self.start > block_number:
            return
        if self.end is not None and self.end < block_number:
            return
        logger.info("Blocks: %i", counter)
        f.write(json.dumps(
            {'blockNumber': block_number, 'balance': balance})+"\n")



from typing import Collection, Iterable, List
from web3 import Web3
from web3.types import BlockData

from eth_tools.logger import logger


class Block:
    """Wrapper of web3.types.BlockData prodiving a ``transactions_count`` property"""

    def __init__(self, data: BlockData):
        self.data = data

    def __getattr__(self, key):
        if key not in self.data:
            raise AttributeError
        value = self.data[key]
        if hasattr(value, "hex"):
            value = value.hex()
        return value

    @property
    def transactions_count(self):
        return len(self.transactions)

    def __repr__(self):
        return "Block(number={number})".format(**self.data)


class BlockIterator:
    """Iterator over Ethereum blocks
    It will lazily fetch each block from ``start_block`` to ``end_block`` inclusive
    using the provided ``web3`` instance.
    """

    def __init__(
        self,
        web3: Web3,
        start_block: int = None,
        end_block: int = None,
        blocks: Collection[int] = None,
        log_interval: int = None,
    ):
        self.web3 = web3
        if blocks is not None:
            assert (
                start_block is None and end_block is None
            ), "blocks is not compatible with start and end block"
            self._blocks = blocks
        else:
            if start_block is None:
                start_block = 0
            if end_block is None:
                end_block = self.web3.eth.blockNumber
            self._blocks = range(start_block, end_block + 1)

        self.blocks_count = len(self._blocks)
        self._blocks_iter = iter(self._blocks)

        self.log_interval = log_interval
        self._processed_count = 0

    def __len__(self):
        return self.blocks_count

    def __iter__(self):
        logger.info("processing %s blocks", self.blocks_count)
        return self

    @property
    def processed_count(self):
        return self._processed_count

    def __next__(self) -> Block:
        block_number = next(self._blocks_iter)
        self._processed_count += 1
        if self.log_interval and self.processed_count % self.log_interval == 0:
            logger.info("%s/%s", self.processed_count, self.blocks_count)
        block = self.web3.eth.getBlock(block_number)
        return Block(block)
